import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets

from ..dialogs.range_editor import RangeEditor
from ..ui.channel_bar_display_widget import Ui_ChannelBarDisplay


class BarWidget(QtWidgets.QWidget):
    def __init__(self, range=(0, 100), over=20, color=None):
        super().__init__()
        self.range = range
        self.over = over
        self.max = self.range[1]
        self.resizeEvent(None)
        self.color = color
        self.value = 50.0

    def resizeEvent(self, event):
        width = self.size().width()
        parts = 6
        while True:
            px = width / parts
            if px >= 50:
                break
            else:
                parts -= 2
                if parts == 4:
                    break

        if isinstance(self.range[0], int):
            self.ticks = [int(e) for e in np.linspace(self.range[0], self.range[1], parts + 1, True).tolist()]
        else:
            self.ticks = np.linspace(self.range[0], self.range[1], parts + 1, True).tolist()

    def setValue(self, value):
        self.value = float(value)
        self.paintEvent(None)

    def set_color(self, color):
        self.color = color
        self.paintEvent(None)

    def paintEvent(self, e):
        qp = QtGui.QPainter()
        qp.begin(self)
        self.drawWidget(qp)
        qp.end()

    def drawWidget(self, qp):
        font = QtGui.QFont("Serif", 7, QtGui.QFont.Weight.Light)
        qp.setFont(font)

        size = self.size()
        w = size.width()
        h = size.height()

        till = int((w / self.max) * self.value)
        full = int((w / self.max) * self.over)

        if self.value >= self.over:
            qp.setPen(QtGui.QColor(self.color))
            qp.setBrush(QtGui.QColor(self.color))
            qp.drawRect(0, 0, full, h)
            qp.setPen(QtGui.QColor(255, 175, 175))
            qp.setBrush(QtGui.QColor(255, 175, 175))
            qp.drawRect(full, 0, till - full, h)

        else:
            qp.setPen(QtGui.QColor(self.color))
            qp.setBrush(QtGui.QColor(self.color))
            qp.drawRect(0, 0, till, h)

        pen = QtGui.QPen(QtGui.QColor(20, 20, 20), 1, QtCore.Qt.PenStyle.SolidLine)

        qp.setPen(pen)
        qp.setBrush(QtCore.Qt.BrushStyle.NoBrush)
        qp.drawRect(0, 0, w - 1, h - 1)

        for j, val in enumerate(self.ticks):
            val_pos = int(val / self.range[1] * w)

            val = str(val) if isinstance(val, int) else f"{val:.3f}"

            qp.drawLine(val_pos, 0, val_pos, 5)
            metrics = qp.fontMetrics().boundingRect(val)
            fw = metrics.width()
            fh = metrics.height()

            x, y = int(val_pos - fw / 2), 7 + fh
            x = max(2, x)
            x = min(x, w - fw - 2)
            qp.drawText(x, y, val)


class ChannelBarDisplay(Ui_ChannelBarDisplay, QtWidgets.QWidget):
    def __init__(
        self,
        uuid,
        value,
        range,
        over,
        color,
        unit="",
        precision=3,
        tooltip="",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        self.color = color
        self._value_prefix = ""
        self._value = value
        self._name = ""

        self.bar = BarWidget(range, over, color=self.color)
        self.layout.addWidget(self.bar)

        self.instant_value = QtWidgets.QLabel("")
        self.instant_value.setMinimumWidth(150)
        font = self.instant_value.font()
        font.setBold(True)
        font.setPointSize(16)
        self.instant_value.setFont(font)
        self.layout.addWidget(self.instant_value)

        self.layout.setStretch(0, 0)
        self.layout.setStretch(1, 0)
        self.layout.setStretch(2, 1)
        self.layout.setStretch(3, 0)

        self.uuid = uuid
        self.ranges = {}
        self.unit = unit.strip()
        self.precision = precision

        self._transparent = True
        self._tooltip = tooltip

        self.color_btn.clicked.connect(self.select_color)

        self.fm = QtGui.QFontMetrics(self.name.font())

        self.setToolTip(self._tooltip or self._name)

        self.kind = "f"
        if self.kind in "SUVui":
            self.fmt = "{}"
        else:
            self.fmt = f"{{:.{self.precision}f}}"

        self.show()

    def set_selected(self, on):
        palette = self.name.palette()
        if on:
            brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
            brush.setStyle(QtCore.Qt.BrushStyle.SolidPattern)
            palette.setBrush(QtGui.QPalette.ColorGroup.Active, QtGui.QPalette.ColorRole.Text, brush)
        else:
            brush = QtGui.QBrush(QtGui.QColor(self.color))
            brush.setStyle(QtCore.Qt.BrushStyle.SolidPattern)
            palette.setBrush(QtGui.QPalette.ColorGroup.Active, QtGui.QPalette.ColorRole.Text, brush)

        self.name.setPalette(palette)

    def set_precision(self, precision):
        if self.kind == "f":
            self.precision = precision
            self.fmt = f"{{:.{self.precision}f}}"

    def mouseDoubleClickEvent(self, event):
        dlg = RangeEditor(self.unit, self.ranges)
        dlg.exec_()
        if dlg.pressed_button == "apply":
            self.ranges = dlg.result

    def select_color(self):
        color = QtWidgets.QColorDialog.getColor(QtGui.QColor(self.color))
        if color.isValid():
            self.set_color(color.name())

        self.bar.set_color(color)

    def set_fmt(self, fmt):
        if self.kind in "SUV":
            self.fmt = "{}"
        elif self.kind == "f":
            self.fmt = f"{{:.{self.precision}f}}"
        else:
            if fmt == "hex":
                self.fmt = "0x{:X}"
            elif fmt == "bin":
                self.fmt = "0b{:b}"
            elif fmt == "phys":
                self.fmt = "{}"

    def set_color(self, color):
        self.color = color
        self.set_name(self._name)
        self.set_value(self._value)
        self.color_btn.setStyleSheet(f"background-color: {color};")

        self.bar.set_color(color)

        palette = self.name.palette()

        brush = QtGui.QBrush(QtGui.QColor(color))
        brush.setStyle(QtCore.Qt.BrushStyle.SolidPattern)
        palette.setBrush(QtGui.QPalette.ColorGroup.Active, QtGui.QPalette.ColorRole.Text, brush)

        self.name.setPalette(palette)
        self.instant_value.setPalette(palette)

    def set_name(self, text=""):
        self.setToolTip(self._tooltip or text)
        self._name = text

    def set_prefix(self, text=""):
        self._value_prefix = text

    def update(self):
        width = self.name.size().width()
        if self.unit:
            self.name.setText(
                self.fm.elidedText(f"{self._name} ({self.unit})", QtCore.Qt.TextElideMode.ElideMiddle, width)
            )
        else:
            self.name.setText(self.fm.elidedText(self._name, QtCore.Qt.TextElideMode.ElideMiddle, width))
        self.set_value(self._value, update=True)

    def set_value(self, value, update=False):
        if self._value == value and update is False:
            return

        self.bar.setValue(value)
        self.instant_value.setText(self.fmt.format(value))

    def keyPressEvent(self, event):
        key = event.key()
        modifier = event.modifiers()
        if modifier == QtCore.Qt.KeyboardModifier.ControlModifier and key == QtCore.Qt.Key.Key_C:
            QtWidgets.QApplication.instance().clipboard().setText(self._name)

        else:
            super().keyPressEvent(event)

    def resizeEvent(self, event):
        width = self.name.size().width()
        if self.unit:
            self.name.setText(
                self.fm.elidedText(f"{self._name} ({self.unit})", QtCore.Qt.TextElideMode.ElideMiddle, width)
            )
        else:
            self.name.setText(self.fm.elidedText(self._name, QtCore.Qt.TextElideMode.ElideMiddle, width))

    def text(self):
        return self._name

    def does_not_exist(self):
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/error.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.color_btn.setIcon(icon)
        self.color_btn.setFlat(True)
        self.color_btn.clicked.disconnect()

    def disconnect_slots(self):
        pass



from copy import deepcopy

import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets

from ..ui.channel_stats import Ui_ChannelStats

MONOSPACE_FONT = None


class ChannelStats(Ui_ChannelStats, QtWidgets.QWidget):
    precision_modified = QtCore.Signal()

    def __init__(self, xunit="s", precision=6, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._settings = QtCore.QSettings()

        self.setupUi(self)

        global MONOSPACE_FONT

        if MONOSPACE_FONT is None:
            families = QtGui.QFontDatabase.families()
            for family in (
                "Consolas",
                "Liberation Mono",
                "DejaVu Sans Mono",
                "Droid Sans Mono",
                "Liberation Mono",
                "Roboto Mono",
                "Monaco",
                "Courier",
            ):
                if family in families:
                    MONOSPACE_FONT = family
                    break

        font = QtGui.QFont(MONOSPACE_FONT)
        self.setFont(font)

        self.xunit = xunit.strip()
        self.color = "#000000"
        self.fmt = "phys"
        self.name_template = '<html><head/><body><p><span style=" font-size:11pt; font-weight:600; color:{};">{}</span></p></body></html>'
        self._name = "Please select a single channel"
        self.name.setStyleSheet("background-color: transparent;")

        for i in range(10):
            label = self.findChild(QtWidgets.QLabel, f"xunit{i}")
            label.setText(f" {self.xunit}")

        self.precision.addItems(["Full float precision"] + [f"{i} float decimals" for i in range(16)])
        self.precision.setCurrentIndex(self._settings.value("stats_float_precision", 6, type=int) + 1)

        self.precision.currentIndexChanged.connect(self.set_float_precision)

    def set_stats(self, stats):
        if not stats:
            self.clear()
            return

        self._stats = deepcopy(stats)
        precision = self._settings.value("stats_float_precision", 6, type=int)
        fmt = f" {{:.{precision}f}}"

        color = stats["color"]
        if stats:
            for name, value in stats.items():
                if name == "unit":
                    for i in range(1, 23):
                        label = self.findChild(QtWidgets.QLabel, f"unit{i}")
                        if label:
                            label.setText(f" {value}")
                    self.selected_gradient_unit.setText(f" {value}/{self.xunit}")
                    self.visible_gradient_unit.setText(f" {value}/{self.xunit}")
                    self.overall_gradient_unit.setText(f" {value}/{self.xunit}")
                    self.selected_integral_unit.setText(f" {value}*{self.xunit}")
                    self.visible_integral_unit.setText(f" {value}*{self.xunit}")
                    self.overall_integral_unit.setText(f" {value}*{self.xunit}")
                elif name == "name":
                    self._name = value
                    self.name.setText(self.name_template.format(self.color, self._name))
                elif name == "color":
                    self.color = value
                    self.name.setText(self.name_template.format(color, self._name))

                elif name in ("region", "color"):
                    continue
                else:
                    label = self.findChild(QtWidgets.QLabel, name)

                    if precision >= 0:
                        if isinstance(value, (float, np.floating)):
                            label.setText(fmt.format(value))
                        else:
                            label.setText(str(value))
                    else:
                        label.setText(str(value))

            if stats["region"]:
                self.cursor_group.setHidden(True)
                self.region_group.setHidden(False)
            else:
                self.cursor_group.setHidden(False)
                self.region_group.setHidden(True)
        else:
            self.clear()

    def clear(self):
        self._name = "Please select a single channel"
        self.color = "#000000"
        self.name.setText(self.name_template.format(self.color, self._name))
        for k, group in enumerate(
            (
                self.cursor_group,
                self.region_group,
                self.visible_group,
                self.overall_group,
            )
        ):
            layout = group.layout()
            rows = layout.rowCount()
            for i in range(rows):
                label = layout.itemAtPosition(i, 1)
                if label is not None:
                    label = label.widget()
                    label.setText("")
                label = layout.itemAtPosition(i, 2)
                if label is not None:
                    label = label.widget()
                    if label.objectName().startswith("unit"):
                        label.setText("")

    def set_float_precision(self, index):
        self._settings.setValue("stats_float_precision", index - 1)
        self.precision_modified.emit()



from functools import partial

import numpy as np
import pyqtgraph as pg
from PySide6 import QtCore, QtGui, QtWidgets

from ...signal import Signal
from ..ui.xy import Ui_XYDisplay

ARROW = pg.arrayToQPath(np.array([0.0, -1.0, 0.0, -1.0, 0.0]), np.array([0.0, 1.0, 0.0, -1.0, 0.0]), connect="all")


class XY(Ui_XYDisplay, QtWidgets.QWidget):
    add_channels_request = QtCore.Signal(list)
    timestamp_changed_signal = QtCore.Signal(object, float)

    def __init__(
        self,
        x_channel=None,
        y_channel=None,
        color="#00ff00",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        self.closed = False
        self.setContentsMargins(0, 0, 0, 0)

        self._settings = QtCore.QSettings()

        self.plot = pg.PlotWidget()
        self.plot_layout.addWidget(self.plot)

        self.plot.plotItem.scene().sigMouseClicked.connect(self.clicked)
        self._pen = color

        # self.curve = self.plot.plot(x=[], y=[], symbol="o", symbolSize=4)
        self.curve = pg.PlotCurveItem(
            size=0,
            pen=self._pen,
            antialias=False,
        )
        self.arrows = pg.ScatterPlotItem(
            size=0,
            pen=self._pen,
            brush=self._pen,
            symbolPen=self._pen,
            symbolBrush=self._pen,
            symbolSize=4,
            antialias=False,
            useCache=False,
        )
        self.plot.addItem(self.curve)
        self.plot.addItem(self.arrows)
        self.marker = self.plot.plot(
            x=[],
            y=[],
            symbol="o",
            symbolPen={"color": self._settings.value("cursor_color", "#ff0000"), "width": 2},
            symbolBrush=QtGui.QBrush(),
            symbolSize=10,
        )

        self.x_search_btn.clicked.connect(partial(self.search, target="x"))
        self.y_search_btn.clicked.connect(partial(self.search, target="y"))

        self.x_channel_edit.editingFinished.connect(partial(self.search, target="x", edit=True))
        self.y_channel_edit.editingFinished.connect(partial(self.search, target="y", edit=True))

        self.show()

        self.timestamp.valueChanged.connect(self._timestamp_changed)
        self.timestamp_slider.valueChanged.connect(self._timestamp_slider_changed)

        self._inhibit = False

        self._x = None
        self._x_interp = None
        self._y = None
        self._y_interp = None
        self._timebase = None
        self._timestamp = None

        self._requested_channel = None

        self.set_x(x_channel)
        self.set_y(y_channel)

        self.update_plot()

    def add_new_channels(self, channels):
        channel = channels[0]
        if self._requested_channel == "x":
            self.set_x(channel)

        elif self._requested_channel == "y":
            self.set_y(channel)

        self._requested_channel = None

        self.update_plot()

    def clicked(self, event):
        scene_pos = event.scenePos()
        pos = self.plot.plotItem.vb.mapSceneToView(scene_pos)
        x = pos.x()
        y = pos.y()

        if self._x is not None and self._y is not None:
            delta = (self._x_interp.samples - x) ** 2 + (self._y_interp.samples - y) ** 2
            idx = np.argmin(delta).flatten()[0]

            x = self._x_interp.samples[idx : idx + 1]
            y = self._y_interp.samples[idx : idx + 1]

            self.marker.setData(
                x=x,
                y=y,
            )

            self.timestamp_slider.setValue(idx)

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()

        if key == QtCore.Qt.Key.Key_C and modifiers == QtCore.Qt.KeyboardModifier.NoModifier:
            event.accept()
            color = QtWidgets.QColorDialog.getColor(self._pen, parent=self)
            if color.isValid():
                self._pen = color.name()
                self.update_plot()

        elif key in (QtCore.Qt.Key.Key_S, QtCore.Qt.Key.Key_F) and modifiers == QtCore.Qt.KeyboardModifier.NoModifier:
            if self._x and self._y and len(self._y):
                min_val, max_val = self._y.samples.min(), self._y.samples.max()
                if min_val == max_val:
                    delta = 1
                else:
                    delta = 0.05 * (max_val - min_val)
                self.plot.setYRange(min_val - delta, max_val + delta, padding=0)

        elif key == QtCore.Qt.Key.Key_W and modifiers == QtCore.Qt.KeyboardModifier.NoModifier:
            event.accept()
            if self._x and self._y and len(self._x):
                min_val, max_val = self._x.samples.min(), self._x.samples.max()
                if min_val == max_val:
                    delta = 1
                else:
                    delta = 0.05 * (max_val - min_val)
                self.plot.setXRange(min_val - delta, max_val + delta, padding=0)
        elif (
            key
            in (
                QtCore.Qt.Key.Key_Left,
                QtCore.Qt.Key.Key_Right,
                QtCore.Qt.Key.Key_PageUp,
                QtCore.Qt.Key.Key_PageDown,
                QtCore.Qt.Key.Key_Home,
                QtCore.Qt.Key.Key_End,
            )
            and modifiers == QtCore.Qt.KeyboardModifier.NoModifier
        ):
            self.timestamp_slider.keyPressEvent(event)

    def search(self, *args, edit=False, target=None, **kwargs):
        self._requested_channel = target
        if edit:
            if target == "x":
                channels = [self.x_channel_edit.text().strip()]
            else:
                channels = [self.y_channel_edit.text().strip()]
        else:
            channels = []

        self.add_channels_request.emit(channels)

    def set_timestamp(self, stamp, emit=True, spinbox=False):
        if stamp is None or self._timebase is None or not len(self._timebase):
            self.marker.setData(x=[], y=[])

        else:
            idx = np.searchsorted(self._timebase, stamp, side="right") - 1

            new_stamp = self._timebase[idx]

            if spinbox:
                if new_stamp == self._timestamp and stamp > new_stamp:
                    idx += 1
                    new_stamp = self._timebase[idx]

            stamp = new_stamp

            x = self._x_interp.samples[idx : idx + 1]
            y = self._y_interp.samples[idx : idx + 1]

            self.marker.setData(
                x=x,
                y=y,
            )

            self._inhibit = True
            self.timestamp_slider.setValue(idx)
            self.timestamp.setValue(stamp)
            self._inhibit = False

            if emit:
                self.timestamp_changed_signal.emit(self, stamp)

        self._timestamp = stamp

    def set_x(self, x):
        if isinstance(x, Signal):
            self._x = x
            self.x_channel_edit.setText(x.name)
            self.plot.plotItem.setLabel("bottom", x.name, x.unit)
        else:
            self._x = None
            self.x_channel_edit.setText("")
            self.plot.plotItem.setLabel("bottom", "", "")

        self.update_plot()
        self.update_timebase()

    def set_y(self, y):
        if isinstance(y, Signal):
            self._y = y
            self.y_channel_edit.setText(y.name)
            self.plot.plotItem.setLabel("left", y.name, y.unit)

        else:
            self._y = None
            self.y_channel_edit.setText("")
            self.plot.plotItem.setLabel("left", "", "")

        self.update_plot()
        self.update_timebase()

    def _timestamp_changed(self, stamp):
        if not self._inhibit:
            self.set_timestamp(stamp, spinbox=True)

    def _timestamp_slider_changed(self, idx):
        if not self._inhibit:
            if not len(self._timebase):
                return

            self.set_timestamp(self._timebase[idx])

    def to_config(self):
        config = {
            "channels": [self._x.name if self._x else "", self._y.name if self._y else ""],
            "color": self._pen,
        }

        return config

    def update_plot(self):
        self.plot.plotItem.getAxis("left").setPen(self._pen)
        self.plot.plotItem.getAxis("left").setTextPen(self._pen)

        x, y = self._x, self._y
        if x is None or y is None:
            self.curve.clear()
            self.arrows.clear()
            self._timebase = None

        elif not len(x) or not len(y):
            self.curve.clear()
            self.arrows.clear()
            self._timebase = None

        else:
            self._timebase = t = np.unique(np.concatenate([x.timestamps, y.timestamps]))
            self._x_interp = x = x.interp(t)
            self._y_interp = y = y.interp(t)

            transform = QtGui.QTransform()

            angles = -np.arctan2(np.diff(y.samples.astype("f8")), np.diff(x.samples.astype("f8"))) * 180 / np.pi

            exit_spots = [
                {
                    "pos": (x.samples[0], y.samples[0]),
                    "symbol": ARROW,
                    "size": 6,
                    "pen": self._pen,
                    "brush": self._pen,
                }
            ]
            for angle, xpos, ypos in zip(angles.tolist(), x.samples[1:].tolist(), y.samples[1:].tolist(), strict=False):
                transform.reset()
                angle_rot = transform.rotate(angle)
                my_rotated_symbol = angle_rot.map(ARROW)

                exit_spots.append(
                    {
                        "pos": (xpos, ypos),
                        "symbol": my_rotated_symbol,
                        "size": 6,
                        "pen": self._pen,
                        "brush": self._pen,
                    }
                )

            # add the spots to the item
            self.arrows.setData(exit_spots)
            self.curve.setData(
                x=x.samples,
                y=y.samples,
                pen=self._pen,
                antialias=False,
            )

            self.set_timestamp(self._timestamp)

    def update_timebase(self):
        if self._timebase is not None and len(self._timebase):
            count = len(self._timebase)
            min_, max_ = self._timebase[0], self._timebase[-1]
            self.timestamp_slider.setRange(0, count - 1)
            if count >= 2:
                self.timestamp.setSingleStep(0.5 * np.min(np.diff(self._timebase)))

        else:
            min_, max_ = 0.0, 0.0
            self.timestamp_slider.setRange(0, 0)

        self.min_t.setText(f"{min_:.9f}s")
        self.max_t.setText(f"{max_:.9f}s")
        self.timestamp.setRange(min_, max_)
        self.set_timestamp(self._timestamp, emit=False)



from typing_extensions import Buffer, Protocol, TypeVar

_T_co = TypeVar("_T_co", covariant=True)


class UnpackFrom(Protocol[_T_co]):
    def __call__(self, buffer: Buffer, offset: int = 0) -> _T_co: ...



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'range_editor_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.7.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QDialog, QGridLayout,
    QListWidgetItem, QPushButton, QSizePolicy, QSpacerItem,
    QWidget)

from asammdf.gui.widgets.list import MinimalListWidget

class Ui_RangeDialog(object):
    def setupUi(self, RangeDialog):
        if not RangeDialog.objectName():
            RangeDialog.setObjectName(u"RangeDialog")
        RangeDialog.resize(1300, 329)
        RangeDialog.setSizeGripEnabled(True)
        RangeDialog.setModal(True)
        self.gridLayout = QGridLayout(RangeDialog)
        self.gridLayout.setSpacing(1)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(1, 1, 1, 1)
        self.ranges = MinimalListWidget(RangeDialog)
        self.ranges.setObjectName(u"ranges")
        self.ranges.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        self.ranges.setDefaultDropAction(Qt.DropAction.MoveAction)

        self.gridLayout.addWidget(self.ranges, 0, 0, 7, 1)

        self.verticalSpacer = QSpacerItem(20, 271, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 4, 1, 1, 1)

        self.reset_btn = QPushButton(RangeDialog)
        self.reset_btn.setObjectName(u"reset_btn")

        self.gridLayout.addWidget(self.reset_btn, 3, 1, 1, 1)

        self.cancel_btn = QPushButton(RangeDialog)
        self.cancel_btn.setObjectName(u"cancel_btn")

        self.gridLayout.addWidget(self.cancel_btn, 6, 1, 1, 1)

        self.apply_btn = QPushButton(RangeDialog)
        self.apply_btn.setObjectName(u"apply_btn")

        self.gridLayout.addWidget(self.apply_btn, 5, 1, 1, 1)

        self.insert_btn = QPushButton(RangeDialog)
        self.insert_btn.setObjectName(u"insert_btn")

        self.gridLayout.addWidget(self.insert_btn, 0, 1, 1, 1)

        self.verticalSpacer_2 = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer_2, 1, 1, 1, 1)

        self.gridLayout.setRowStretch(4, 1)
        self.gridLayout.setColumnStretch(0, 1)
        QWidget.setTabOrder(self.ranges, self.insert_btn)
        QWidget.setTabOrder(self.insert_btn, self.reset_btn)
        QWidget.setTabOrder(self.reset_btn, self.apply_btn)
        QWidget.setTabOrder(self.apply_btn, self.cancel_btn)

        self.retranslateUi(RangeDialog)

        QMetaObject.connectSlotsByName(RangeDialog)
    # setupUi

    def retranslateUi(self, RangeDialog):
        RangeDialog.setWindowTitle(QCoreApplication.translate("RangeDialog", u"Edit value range colors", None))
        self.reset_btn.setText(QCoreApplication.translate("RangeDialog", u"Reset", None))
        self.cancel_btn.setText(QCoreApplication.translate("RangeDialog", u"Cancel", None))
        self.apply_btn.setText(QCoreApplication.translate("RangeDialog", u"Apply", None))
        self.insert_btn.setText(QCoreApplication.translate("RangeDialog", u"Insert range", None))
    # retranslateUi



from enum import IntEnum
import os
from typing import Final, Literal

from typing_extensions import Any, TypedDict

from .types import StrPath


class IntegerInterpolation(IntEnum):
    REPEAT_PREVIOUS_SAMPLE = 0
    LINEAR_INTERPOLATION = 1
    HYBRID_INTERPOLATION = 2


class FloatInterpolation(IntEnum):
    REPEAT_PREVIOUS_SAMPLE = 0
    LINEAR_INTERPOLATION = 1


class _GlobalOptions(TypedDict):
    read_fragment_size: int
    write_fragment_size: int
    use_display_names: bool
    single_bit_uint_as_bool: bool
    integer_interpolation: IntegerInterpolation
    float_interpolation: FloatInterpolation
    copy_on_get: bool
    temporary_folder: StrPath | None
    raise_on_multiple_occurrences: bool
    fill_0_for_missing_computation_channels: bool


GLOBAL_OPTIONS: Final[_GlobalOptions] = {
    "read_fragment_size": 256 * 1024 * 1024,
    "write_fragment_size": 4 * 1024 * 1024,
    "use_display_names": True,
    "single_bit_uint_as_bool": False,
    "integer_interpolation": IntegerInterpolation.REPEAT_PREVIOUS_SAMPLE,
    "float_interpolation": FloatInterpolation.LINEAR_INTERPOLATION,
    "copy_on_get": True,
    "temporary_folder": None,
    "raise_on_multiple_occurrences": True,
    "fill_0_for_missing_computation_channels": False,
}

_Opt = Literal[
    "read_fragment_size",
    "write_fragment_size",
    "use_display_names",
    "single_bit_uint_as_bool",
    "integer_interpolation",
    "float_interpolation",
    "copy_on_get",
    "temporary_folder",
    "raise_on_multiple_occurrences",
    "fill_0_for_missing_computation_channels",
]


def set_global_option(opt: _Opt, value: Any) -> None:
    if opt not in GLOBAL_OPTIONS:
        raise KeyError(f'Unknown global option "{opt}"')

    if opt == "read_fragment_size":
        GLOBAL_OPTIONS[opt] = int(value)
    elif opt == "write_fragment_size":
        GLOBAL_OPTIONS[opt] = min(int(value), 4 * 1024 * 1024)
    elif opt in (
        "use_display_names",
        "single_bit_uint_as_bool",
        "copy_on_get",
        "raise_on_multiple_occurrences",
        "fill_0_for_missing_computation_channels",
    ):
        GLOBAL_OPTIONS[opt] = bool(value)
    elif opt == "integer_interpolation":
        GLOBAL_OPTIONS[opt] = IntegerInterpolation(value)
    elif opt == "float_interpolation":
        GLOBAL_OPTIONS[opt] = FloatInterpolation(value)
    elif opt == "temporary_folder":
        value = value or None
        if value is not None:
            os.makedirs(value, exist_ok=True)
        GLOBAL_OPTIONS[opt] = value


def get_global_option(opt: _Opt) -> object:
    return GLOBAL_OPTIONS[opt]


def get_option(opt: _Opt, instance_options: _GlobalOptions) -> object:
    value: object = instance_options[opt]
    if value is None:
        value = get_global_option(opt)

    return value



#!/usr/bin/env python

"""
Main test function to execute all tests found in the current directory
"""

from pathlib import Path
import sys
import unittest

import xmlrunner


def main() -> bool:
    tests = unittest.TestLoader().discover(".", "test_*.py")
    testResult = xmlrunner.XMLTestRunner(output=str(Path(".").resolve() / "test-reports")).run(tests)

    return not testResult.wasSuccessful()


if __name__ == "__main__":
    sys.exit(main())



from collections.abc import Collection
from copy import deepcopy
import json
from pathlib import Path
import re
from traceback import format_exc
import typing
from typing import Final, Literal, Union

import lxml.etree
from pyqtgraph import functions as fn
from PySide6 import QtCore, QtGui
from typing_extensions import NotRequired, TypedDict

from asammdf.blocks import v4_constants as v4c
from asammdf.blocks.utils import SignalFlags

C_FUNCTION: Final = re.compile(r"\s+(?P<function>\S+)\s*\(\s*struct\s+DATA\s+\*data\s*\)")
COLOR_MAPS: Final = {
    "Accent": ["#7fc97f", "#beaed4", "#fdc086", "#ffff99", "#386cb0", "#f0027f", "#bf5b16", "#666666"],
    "Dark2": ["#1b9e77", "#d95f02", "#7570b3", "#e7298a", "#66a61e", "#e6ab02", "#a6761d", "#666666"],
    "Paired": [
        "#a6cee3",
        "#1f78b4",
        "#b2df8a",
        "#33a02c",
        "#fb9a99",
        "#e31a1c",
        "#fdbf6f",
        "#ff7f00",
        "#cab2d6",
        "#6a3d9a",
        "#ffff99",
        "#b15928",
    ],
    "Pastel1": ["#fbb4ae", "#b3cde3", "#ccebc5", "#decbe4", "#fed9a6", "#ffffcc", "#e5d8bd", "#fddaec", "#f2f2f2"],
    "Pastel2": ["#b3e2cd", "#fdcdac", "#cbd5e8", "#f4cae4", "#e6f5c9", "#fff2ae", "#f1e2cc", "#cccccc"],
    "Set1": ["#e41a1c", "#377eb8", "#4daf4a", "#984ea3", "#ff7f00", "#ffff33", "#a65628", "#f781bf", "#999999"],
    "Set2": ["#66c2a5", "#fc8d62", "#8da0cb", "#e78ac3", "#a6d854", "#ffd92f", "#e5c494", "#b3b3b3"],
    "Set3": [
        "#8dd3c7",
        "#ffffb3",
        "#bebada",
        "#fb8072",
        "#80b1d3",
        "#fdb462",
        "#b3de69",
        "#fccde5",
        "#d9d9d9",
        "#bc80bd",
        "#ccebc5",
        "#ffed6f",
    ],
    "tab10": [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
    ],
    "tab20": [
        "#1f77b4",
        "#aec7e8",
        "#ff7f0e",
        "#ffbb78",
        "#2ca02c",
        "#98df8a",
        "#d62728",
        "#ff9896",
        "#9467bd",
        "#c5b0d5",
        "#8c564b",
        "#c49c94",
        "#e377c2",
        "#f7b6d2",
        "#7f7f7f",
        "#c7c7c7",
        "#bcbd22",
        "#dbdb8d",
        "#17becf",
        "#9edae5",
    ],
    "tab20b": [
        "#393b79",
        "#5254a3",
        "#6b6ecf",
        "#9c9ede",
        "#637939",
        "#8ca252",
        "#b5cf6b",
        "#cedb9c",
        "#8c6d31",
        "#bd9e39",
        "#e7ba52",
        "#e7cb94",
        "#843c39",
        "#ad494a",
        "#d6616b",
        "#e7969c",
        "#7b4173",
        "#a55194",
        "#ce6dbd",
        "#de9ed6",
    ],
    "tab20c": [
        "#3182bd",
        "#6baed6",
        "#9ecae1",
        "#c6dbef",
        "#e6550d",
        "#fd8d3c",
        "#fdae6b",
        "#fdd0a2",
        "#31a354",
        "#74c476",
        "#a1d99b",
        "#c7e9c0",
        "#756bb1",
        "#9e9ac8",
        "#bcbddc",
        "#dadaeb",
        "#636363",
        "#969696",
        "#bdbdbd",
        "#d9d9d9",
    ],
}
COLORS: Final = COLOR_MAPS["tab10"]
COLORS_COUNT: Final = len(COLORS)
COMPARISON_NAME: Final = re.compile(r"(\s*\d+:)?(?P<name>.+)")


class _ChannelBaseDict(TypedDict):
    color: str
    comment: str | None
    common_axis: bool
    enabled: bool
    flags: int
    fmt: str
    individual_axis: bool
    name: str
    origin_uuid: str
    precision: int
    ranges: list[dict[str, object]]
    type: Literal["channel"]
    unit: str


class _ChannelNotComputedDict(_ChannelBaseDict):
    computed: Literal[False]
    conversion: NotRequired[dict[str, object]]
    mode: Literal["phys"]
    y_range: list[float]


class _ChannelComputedDict(_ChannelBaseDict):
    computed: Literal[True]
    computation: dict[str, object]
    conversion: object
    user_defined_name: str | None


_ChannelDict = _ChannelComputedDict | _ChannelNotComputedDict


class _ChannelGroupDict(TypedDict):
    channels: list[Union["_ChannelGroupDict", _ChannelDict]]
    enabled: bool
    name: str | None
    origin_uuid: str
    pattern: dict[str, object] | None
    ranges: list[dict[str, object]]
    type: Literal["group"]


def extract_mime_names(data: QtCore.QMimeData, disable_new_channels: bool | None = None) -> list[str]:
    def fix_comparison_name(
        data: list[_ChannelGroupDict | _ChannelDict], disable_new_channels: bool | None = None
    ) -> None:
        for item in data:
            if item["type"] == "channel":
                if disable_new_channels is not None:
                    item["enabled"] = not disable_new_channels

                if (
                    item["group_index"],  # type: ignore[typeddict-item]
                    item["channel_index"],  # type: ignore[typeddict-item]
                ) != (-1, -1):
                    match = COMPARISON_NAME.match(item["name"])
                    if match is None:
                        raise RuntimeError(f"cannot parse '{item['name']}'")
                    name = match.group("name").strip()
                    item["name"] = name
            else:
                if disable_new_channels is not None:
                    item["enabled"] = not disable_new_channels
                fix_comparison_name(item["channels"], disable_new_channels=disable_new_channels)

    names: list[str] = []
    if data.hasFormat("application/octet-stream-asammdf"):
        data_data = data.data("application/octet-stream-asammdf").data()
        data_bytes = data_data.tobytes() if isinstance(data_data, memoryview) else data_data
        text = data_bytes.decode("utf-8")
        obj = json.loads(text, cls=ExtendedJsonDecoder)
        fix_comparison_name(obj, disable_new_channels=disable_new_channels)
        names = obj

    return names


def set_mime_enable(mime: list[_ChannelGroupDict | _ChannelDict], enable: bool) -> None:
    for item in mime:
        if item["type"] == "channel":
            item["enabled"] = enable
        else:
            set_mime_enable(item["channels"], enable)


def flatten_dsp(channels: list[_ChannelGroupDict | _ChannelDict]) -> list[str]:
    res: list[str] = []

    for item in channels:
        if item["type"] == "group":
            res.extend(flatten_dsp(item["channels"]))
        else:
            res.append(item["name"])

    return res


def load_dsp(
    file: Path, background: str = "#000000", flat: bool = False, colors_as_string: bool = False
) -> dict[str, object] | list[str]:
    if not colors_as_string and isinstance(background, str):
        background = fn.mkColor(background)

    def parse_conversions(display: lxml.etree._Element | None) -> dict[str | None, dict[str, object]]:
        conversions: dict[str | None, dict[str, object]] = {}

        if display is None:
            return conversions

        for item in display.findall("COMPU_METHOD"):
            try:
                name = item.get("name")
                conv: dict[str, object] = {
                    "name": name,
                    "comment": item.get("description"),
                    "unit": item.get("unit"),
                }

                conversion_type = int(item.attrib["cnv_type"])
                match conversion_type:
                    case 0:
                        conv["conversion_type"] = v4c.CONVERSION_TYPE_LIN

                        coeffs = item.find("COEFFS_LINIAR")

                        if coeffs is None:
                            raise RuntimeError("cannot find 'COEFFS_LINIAR' element")

                        conv["a"] = float(coeffs.attrib["P1"])
                        conv["b"] = float(coeffs.attrib["P2"])

                    case 9:
                        conv["conversion_type"] = v4c.CONVERSION_TYPE_RAT

                        coeffs = item.find("COEFFS")

                        if coeffs is None:
                            raise RuntimeError("cannot find 'COEFFS' element")

                        for i in range(1, 7):
                            conv[f"P{i}"] = float(coeffs.attrib[f"P{i}"])

                    case 11:
                        conv["conversion_type"] = v4c.CONVERSION_TYPE_TABX
                        vtab = item.find("COMPU_VTAB")

                        if vtab is not None:
                            for i, item in enumerate(vtab.findall("tab")):
                                conv[f"val_{i}"] = float(item.attrib["min"])
                                text = item.get("text")
                                if isinstance(text, bytes):
                                    text = text.decode("utf-8", errors="replace")
                                conv[f"text_{i}"] = text

                    case 12:
                        conv["conversion_type"] = v4c.CONVERSION_TYPE_RTABX
                        vtab = item.find("COMPU_VTAB_RANGE")

                        if vtab is not None:
                            text = vtab.get("default")
                            if isinstance(text, bytes):
                                text = text.decode("utf-8", errors="replace")
                            conv["default_addr"] = vtab.get("default")
                            for i, item in enumerate(vtab.findall("tab_range")):
                                conv[f"upper_{i}"] = float(item.attrib["max"])
                                conv[f"lower_{i}"] = float(item.attrib["min"])
                                text = item.get("text")
                                if isinstance(text, bytes):
                                    text = text.decode("utf-8", errors="replace")
                                conv[f"text_{i}"] = text
                    case _:
                        continue

                conversions[name] = conv

            except:
                print(format_exc())
                continue

        return conversions

    def parse_channels(
        display: lxml.etree._Element, conversions: dict[str | None, dict[str, object]]
    ) -> list[_ChannelGroupDict | _ChannelDict]:
        channels: list[_ChannelGroupDict | _ChannelDict] = []
        for elem in display.iterchildren():
            if elem.tag == "CHANNEL":
                channel_name = elem.attrib["name"] or "unnamed"

                comment_elem = elem.find("COMMENT")
                if comment_elem is not None:
                    comment = elem.get("text")
                else:
                    comment = ""

                color_ = int(elem.attrib["color"])
                c = 0
                for i in range(3):
                    c = c << 8
                    c += color_ & 0xFF
                    color_ = color_ >> 8

                ch_color = c

                gain = abs(float(elem.attrib["gain"]))
                offset = float(elem.attrib["offset"]) / 100

                multi_color = elem.find("MULTI_COLOR")

                ranges: list[dict[str, object]] = []

                if multi_color is not None:
                    for color in multi_color.findall("color"):
                        some_elem = color.find("min")
                        if some_elem is None:
                            raise RuntimeError("cannot find element 'min'")
                        min_ = float(some_elem.attrib["data"])
                        some_elem = color.find("max")
                        if some_elem is None:
                            raise RuntimeError("cannot find element 'max'")
                        max_ = float(some_elem.attrib["data"])
                        some_elem = color.find("color")
                        if some_elem is None:
                            raise RuntimeError("cannot find element 'color'")
                        color_ = int(some_elem.attrib["data"])
                        c = 0
                        for i in range(3):
                            c = c << 8
                            c += color_ & 0xFF
                            color_ = color_ >> 8
                        font_color = f"#{c:06X}" if colors_as_string else fn.mkColor(f"#{c:06X}")
                        ranges.append(
                            {
                                "background_color": background,
                                "font_color": font_color,
                                "op1": "<=",
                                "op2": "<=",
                                "value1": min_,
                                "value2": max_,
                            }
                        )

                chan: _ChannelNotComputedDict = {
                    "color": f"#{ch_color:06X}",
                    "common_axis": False,
                    "computed": False,
                    "flags": 0,
                    "comment": comment,
                    "enabled": elem.get("on") == "1" and elem.get("trc_fmt") != "2",
                    "fmt": "{}",
                    "individual_axis": False,
                    "name": channel_name,
                    "mode": "phys",
                    "precision": 3,
                    "ranges": ranges,
                    "unit": "",
                    "type": "channel",
                    "y_range": sorted(
                        [
                            -gain * offset,
                            -gain * offset + 19 * gain,
                        ]
                    ),
                    "origin_uuid": "000000000000",
                }

                conv_name = elem.get("cnv_name")
                if conv_name in conversions:
                    chan["conversion"] = deepcopy(conversions[conv_name])

                channels.append(chan)

            elif str(elem.tag).startswith("GROUP"):
                channels.append(
                    {
                        "name": elem.get("data") or "unnamed",
                        "enabled": elem.get("on") == "1",
                        "type": "group",
                        "channels": parse_channels(elem, conversions=conversions),
                        "pattern": None,
                        "origin_uuid": "000000000000",
                        "ranges": [],
                    }
                )

            elif elem.tag == "CHANNEL_PATTERN":
                try:
                    filter_type = elem.get("filter_type")
                    filter_value: float
                    if filter_type in ("None", None):
                        filter_type = "Unspecified"
                        filter_value = 0
                        raw = False
                    else:
                        filter_value = float(elem.attrib["filter_value"])
                        raw = bool(int(elem.attrib["filter_use_raw"]))

                    info: dict[str, object] = {
                        "pattern": elem.get("name_pattern"),
                        "name": elem.get("name_pattern") or "unnamed",
                        "match_type": "Wildcard",
                        "filter_type": filter_type,
                        "filter_value": filter_value,
                        "raw": raw,
                    }

                    multi_color = elem.find("MULTI_COLOR")

                    ranges = []

                    if multi_color is not None:
                        for color in multi_color.findall("color"):
                            some_elem = color.find("min")
                            if some_elem is None:
                                raise RuntimeError("cannot find element 'min'")
                            min_ = float(some_elem.attrib["data"])
                            some_elem = color.find("max")
                            if some_elem is None:
                                raise RuntimeError("cannot find element 'max'")
                            max_ = float(some_elem.attrib["data"])
                            some_elem = color.find("color")
                            if some_elem is None:
                                raise RuntimeError("cannot find element 'color'")
                            color_ = int(some_elem.attrib["data"])
                            c = 0
                            for i in range(3):
                                c = c << 8
                                c += color_ & 0xFF
                                color_ = color_ >> 8
                            font_color = f"#{c:06X}" if colors_as_string else fn.mkColor(f"#{c:06X}")
                            ranges.append(
                                {
                                    "background_color": background,
                                    "font_color": font_color,
                                    "op1": "<=",
                                    "op2": "<=",
                                    "value1": min_,
                                    "value2": max_,
                                }
                            )

                    info["ranges"] = ranges

                    channels.append(
                        {
                            "channels": [],
                            "enabled": True,
                            "name": typing.cast(str, info["pattern"]),
                            "pattern": info,
                            "type": "group",
                            "ranges": [],
                            "origin_uuid": "000000000000",
                        }
                    )

                except:
                    print(format_exc())
                    continue

        return channels

    def parse_virtual_channels(display: lxml.etree._Element | None) -> dict[str | None, dict[str, object]]:
        channels: dict[str | None, dict[str, object]] = {}

        if display is None:
            return channels

        for item in display.findall("V_CHAN"):
            try:
                virtual_channel: dict[str, object] = {}

                parent = item.find("VIR_TIME_CHAN")
                vtab = item.find("COMPU_VTAB")
                if parent is None or vtab is None:
                    continue

                name = item.get("name")

                virtual_channel["name"] = name
                virtual_channel["parent"] = parent.attrib["data"]
                elem = item.find("description")
                if elem is None:
                    raise RuntimeError("cannot find element 'description'")
                virtual_channel["comment"] = elem.attrib["data"]

                conv: dict[str, object] = {}
                for i, item in enumerate(vtab.findall("tab")):
                    conv[f"val_{i}"] = float(item.attrib["min"])
                    text = item.get("text")
                    if isinstance(text, bytes):
                        text = text.decode("utf-8", errors="replace")
                    conv[f"text_{i}"] = text

                virtual_channel["vtab"] = conv

                channels[name] = virtual_channel
            except:
                continue

        return channels

    def parse_c_functions(display: lxml.etree._Element | None) -> Collection[str]:
        c_functions: set[str] = set()

        if display is None:
            return c_functions

        for item in display.findall("CALC_FUNC"):
            string = item.text

            if string is None:
                raise RuntimeError("element text is None")

            for match in C_FUNCTION.finditer(string):
                c_functions.add(match.group("function"))

        return sorted(c_functions)

    dsp = lxml.etree.fromstring(Path(file).read_bytes().replace(b"\0", b""), parser=lxml.etree.XMLParser(recover=True))

    conversions = parse_conversions(dsp.find("COMPU_METHODS"))

    elem = dsp.find("DISPLAY_INFO")

    if elem is None:
        raise RuntimeError("cannot find element 'DISPLAY_INFO'")

    channels = parse_channels(elem, conversions)
    c_functions = parse_c_functions(dsp)

    functions: dict[str, object] = {}
    virtual_channels: list[_ChannelGroupDict | _ChannelDict] = []

    for i, ch in enumerate(parse_virtual_channels(dsp.find("VIRTUAL_CHANNEL")).values()):
        virtual_channels.append(
            {
                "color": COLORS[i % len(COLORS)],
                "common_axis": False,
                "computed": True,
                "computation": {
                    "args": {"arg1": []},
                    "type": "python_function",
                    "channel_comment": ch["comment"],
                    "channel_name": ch["name"],
                    "channel_unit": "",
                    "function": f"f_{ch['name']}",
                    "triggering": "triggering_on_all",
                    "triggering_value": "all",
                },
                "flags": int(SignalFlags.computed | SignalFlags.user_defined_conversion),
                "enabled": True,
                "fmt": "{}",
                "individual_axis": False,
                "name": typing.cast(str, ch["parent"]),
                "precision": 3,
                "ranges": [],
                "unit": "",
                "conversion": ch["vtab"],
                "user_defined_name": typing.cast(str | None, ch["name"]),
                "comment": f"Datalyser virtual channel: {ch['comment']}",
                "origin_uuid": "000000000000",
                "type": "channel",
            }
        )

        functions[f"f_{ch['name']}"] = f"def f_{ch['name']}(arg1=0, t=0):\n    return arg1"

    if virtual_channels:
        channels.append(
            {
                "name": "Datalyser Virtual Channels",
                "enabled": False,
                "type": "group",
                "channels": virtual_channels,
                "pattern": None,
                "origin_uuid": "000000000000",
                "ranges": [],
            }
        )

    windows: list[dict[str, object]] = []
    info: dict[str, object] | list[str] = {
        "selected_channels": [],
        "windows": windows,
        "has_virtual_channels": bool(virtual_channels),
        "c_functions": c_functions,
        "functions": functions,
    }

    if flat:
        info = flatten_dsp(channels)
    else:
        plot: dict[str, object] = {
            "type": "Plot",
            "title": "Display channels",
            "maximized": True,
            "minimized": False,
            "configuration": {
                "channels": channels,
                "locked": True,
                "pattern": {},
            },
        }

        windows.append(plot)

    return info


class ExtendedJsonDecoder(json.JSONDecoder):
    def __init__(self, **kwargs):
        kwargs["object_hook"] = self.object_hook
        super().__init__(**kwargs)

    def object_hook(self, obj):
        for key in ("color", "background_color", "font_color"):
            if key in obj:
                obj[key] = fn.mkColor(obj[key])
        return obj


class ExtendedJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, QtGui.QBrush):
            return obj.color().name()
        elif isinstance(obj, QtGui.QColor):
            return obj.name()
        else:
            return super().default(obj)


def load_lab(file: Path) -> dict[str, list[str]]:
    sections: dict[str, list[str]] = {}
    with open(file) as lab:
        for line in lab:
            line = line.strip()
            if not line:
                continue

            if line.startswith("[") and line.endswith("]"):
                section_name = line.strip("[]")
                s: list[str] = []
                sections[section_name] = s

            else:
                if "s" in locals():
                    s.append(line)

    return {name: channels for name, channels in sections.items() if channels if name != "SETTINGS"}


def load_channel_names_from_file(file_name: str, lab_section: str = "") -> list[str]:
    file_path = Path(file_name)
    channels: Collection[str]
    extension = file_path.suffix.lower()
    match extension:
        case ".dsp":
            channels = load_dsp(file_path, flat=True)

        case ".dspf":
            with open(file_path) as infile:
                info = json.load(infile, cls=ExtendedJsonDecoder)

            channels = []
            for window in info["windows"]:
                if window["type"] == "Plot":
                    channels.extend(flatten_dsp(window["configuration"]["channels"]))
                elif window["type"] == "Numeric":
                    channels.extend([item["name"] for item in window["configuration"]["channels"]])
                elif window["type"] == "Tabular":
                    channels.extend(window["configuration"]["channels"])

        case ".lab":
            info = load_lab(file_path)
            if info:
                if len(info) > 1 and lab_section:
                    channels = info[lab_section]
                else:
                    channels = list(info.values())[0]

                channels = [name.split(";")[0] for name in channels]

        case ".cfg":
            with open(file_path) as infile:
                info = json.load(infile, cls=ExtendedJsonDecoder)
            channels = info.get("selected_channels", [])
        case ".txt":
            try:
                with open(file_path) as infile:
                    info = json.load(infile, cls=ExtendedJsonDecoder)
                channels = info.get("selected_channels", [])
            except:
                with open(file_path) as infile:
                    channels = [line.strip() for line in infile.readlines()]
                    channels = [name for name in channels if name]

    return sorted(set(channels))



import os

from PySide6 import QtCore, QtGui, QtWidgets

DEFAULT_TIMEOUT = int(os.environ.get("ASAMMDF_ERROR_DIALOG_TIMEOUT", "60"))


class MessageBox(QtWidgets.QMessageBox):
    def __init__(self, *args, **kwargs):
        self.timeout = kwargs.pop("timeout", DEFAULT_TIMEOUT)
        informative_text = kwargs.pop("informative_text", "")
        detailed_text = kwargs.pop("detailed_text", "")
        escapeButton = kwargs.pop("escapeButton", None)
        defaultButton = kwargs.pop("defaultButton", None)
        markdown = kwargs.pop("markdown", False)

        super().__init__(*args, **kwargs)

        self.label = label = self.findChild(QtWidgets.QLabel, "qt_msgbox_label")
        idx = self.layout().indexOf(label)
        position = self.layout().getItemPosition(idx)
        self.layout().removeWidget(label)

        layout = QtWidgets.QVBoxLayout()
        layout.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetMinAndMaxSize)

        self.scroll_contents = scroll_contents = QtWidgets.QWidget()
        scroll_contents.setLayout(layout)

        self.scroll = scroll = QtWidgets.QScrollArea()
        scroll.setWidgetResizable(False)
        scroll.setWidget(scroll_contents)
        scroll.setFrameShape(QtWidgets.QFrame.Shape.NoFrame)

        layout.addWidget(label)

        self.layout().addWidget(scroll, *position)

        self.original_text = self.text()
        if markdown:
            self.setTextFormat(QtCore.Qt.TextFormat.MarkdownText)

        if defaultButton is not None:
            self.setDefaultButton(defaultButton)

        if self.defaultButton() is not None:
            if markdown:
                self.setText(
                    f"""{self.original_text}

* * *

This message will be closed in {self.timeout}s
* Default button - [{self.defaultButton().text().strip("&")}]
* Abort the countdown - [F1]"""
                )
            else:
                self.setText(
                    f"{self.original_text}\n\nThis message will be closed in {self.timeout}s\n"
                    f"Default button - [{self.defaultButton().text().strip('&')}]\n"
                    "Abort the countdown - [F1]"
                )
        else:
            if markdown:
                self.setText(
                    f"""{self.original_text}

* * *

This message will be closed in {self.timeout}s
* Abort the countdown - [F1]"""
                )

            else:
                self.setText(
                    f"{self.original_text}\n\nThis message will be closed in {self.timeout}s\n"
                    "Abort the countdown - [F1]"
                )

        if escapeButton is not None:
            self.setEscapeButton(escapeButton)

        self.show()

        if informative_text:
            self.setInformativeText(informative_text)

        if detailed_text:
            self.setDetailedText(detailed_text)

        self.scroll.setMinimumWidth(min(800, self.scroll_contents.width()))
        self.scroll.setMinimumHeight(min(800, self.scroll_contents.height()))

        rect = QtCore.QRect(0, 0, self.scroll_contents.width() + 70, self.scroll_contents.height() + 90)

        cursor = QtGui.QCursor()
        for screen in QtWidgets.QApplication.screens():
            pos = cursor.pos(screen)
            screen_rect = screen.availableGeometry()
            if screen_rect.contains(pos):
                rect.moveCenter(screen_rect.center())
                break
        else:
            screen_rect = screen.availableGeometry()
            rect.moveCenter(screen_rect.center())

        self.setGeometry(rect)

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.tick)
        self.timer.start(1000)

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key.Key_F1:
            self.timer.stop()
            self.setText(self.original_text)
            event.accept()
        else:
            super().keyPressEvent(event)

    def tick(self):
        self.timeout -= 1
        default = self.defaultButton()

        if self.timeout <= 0:
            self.timer.stop()
            if default is not None:
                default.animateClick()
            else:
                self.done(0)
        else:
            if self.defaultButton() is not None:
                if self.textFormat() == QtCore.Qt.TextFormat.MarkdownText:
                    self.setText(
                        f"""{self.original_text}

* * *

This message will be closed in {self.timeout}s
* Default button - [{self.defaultButton().text().strip("&")}]
* Abort the countdown - [F1]"""
                    )
                else:
                    self.setText(
                        f"{self.original_text}\n\nThis message will be closed in {self.timeout}s\n"
                        f"Default button - [{self.defaultButton().text().strip('&')}]\n"
                        "Abort the countdown - [F1]"
                    )
            else:
                if self.textFormat() == QtCore.Qt.TextFormat.MarkdownText:
                    self.setText(
                        f"""{self.original_text}

* * *

This message will be closed in {self.timeout}s
* Abort the countdown - [F1]"""
                    )

                else:
                    self.setText(
                        f"{self.original_text}\n\nThis message will be closed in {self.timeout}s\n"
                        "Abort the countdown - [F1]"
                    )

    @classmethod
    def about(
        cls,
        parent,
        title,
        text,
        buttons=QtWidgets.QMessageBox.StandardButton.Ok,
        defaultButton=QtWidgets.QMessageBox.StandardButton.Ok,
        escapeButton=QtWidgets.QMessageBox.StandardButton.Ok,
        timeout=DEFAULT_TIMEOUT,
        markdown=False,
        informative_text="",
        detailed_text="",
    ):
        msg = cls(
            QtWidgets.QMessageBox.Icon.NoIcon,
            title,
            text,
            buttons,
            parent,
            timeout=timeout,
            defaultButton=defaultButton,
            escapeButton=escapeButton,
            markdown=markdown,
            informative_text=informative_text,
            detailed_text=detailed_text,
        )

        return msg.exec()

    @classmethod
    def critical(
        cls,
        parent,
        title,
        text,
        buttons=QtWidgets.QMessageBox.StandardButton.Ok,
        defaultButton=QtWidgets.QMessageBox.StandardButton.Ok,
        escapeButton=QtWidgets.QMessageBox.StandardButton.Ok,
        timeout=DEFAULT_TIMEOUT,
        markdown=False,
        informative_text="",
        detailed_text="",
    ):
        msg = cls(
            QtWidgets.QMessageBox.Icon.Critical,
            title,
            text,
            buttons,
            parent,
            timeout=timeout,
            defaultButton=defaultButton,
            escapeButton=escapeButton,
            markdown=markdown,
            informative_text=informative_text,
            detailed_text=detailed_text,
        )

        return msg.exec()

    @classmethod
    def information(
        cls,
        parent,
        title,
        text,
        buttons=QtWidgets.QMessageBox.StandardButton.Ok,
        defaultButton=QtWidgets.QMessageBox.StandardButton.Ok,
        escapeButton=QtWidgets.QMessageBox.StandardButton.Ok,
        timeout=DEFAULT_TIMEOUT,
        markdown=False,
        informative_text="",
        detailed_text="",
    ):
        msg = cls(
            QtWidgets.QMessageBox.Icon.Information,
            title,
            text,
            buttons,
            parent,
            timeout=timeout,
            defaultButton=defaultButton,
            escapeButton=escapeButton,
            markdown=markdown,
            informative_text=informative_text,
            detailed_text=detailed_text,
        )
        return msg.exec()

    @classmethod
    def question(
        cls,
        parent,
        title,
        text,
        buttons=QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No,
        defaultButton=QtWidgets.QMessageBox.StandardButton.No,
        escapeButton=QtWidgets.QMessageBox.StandardButton.No,
        timeout=DEFAULT_TIMEOUT,
        markdown=False,
        informative_text="",
        detailed_text="",
    ):
        msg = cls(
            QtWidgets.QMessageBox.Icon.Question,
            title,
            text,
            buttons,
            parent,
            timeout=timeout,
            defaultButton=defaultButton,
            escapeButton=escapeButton,
            markdown=markdown,
            informative_text=informative_text,
            detailed_text=detailed_text,
        )
        return msg.exec()

    @classmethod
    def warning(
        cls,
        parent,
        title,
        text,
        buttons=QtWidgets.QMessageBox.StandardButton.Ok,
        defaultButton=QtWidgets.QMessageBox.StandardButton.Ok,
        escapeButton=QtWidgets.QMessageBox.StandardButton.Ok,
        timeout=DEFAULT_TIMEOUT,
        markdown=False,
        informative_text="",
        detailed_text="",
    ):
        msg = cls(
            QtWidgets.QMessageBox.Icon.Warning,
            title,
            text,
            buttons,
            parent,
            timeout=timeout,
            defaultButton=defaultButton,
            escapeButton=escapeButton,
            markdown=markdown,
            informative_text=informative_text,
            detailed_text=detailed_text,
        )
        return msg.exec()



#!/usr/bin/env python
from io import BytesIO
from pathlib import Path
import random
import tempfile
import unittest
import urllib
import urllib.request
from zipfile import ZipFile

import numexpr
import numpy as np
import numpy.typing as npt
from pandas import DataFrame
from typing_extensions import Any

from asammdf import MDF, Signal, SUPPORTED_VERSIONS
from asammdf.blocks.utils import MdfException
from asammdf.mdf import SearchMode

from .utils import cycles, generate_arrays_test_file, generate_test_file

SUPPORTED_VERSIONS = tuple(version for version in SUPPORTED_VERSIONS if "4.20" > version >= "3.20")  # type: ignore[misc]

CHANNEL_LEN = 100000


class TestMDF(unittest.TestCase):
    tempdir_demo: tempfile.TemporaryDirectory[str]
    tempdir_array: tempfile.TemporaryDirectory[str]
    tempdir_general: tempfile.TemporaryDirectory[str]
    tempdir: tempfile.TemporaryDirectory[str]

    def test_measurement(self) -> None:
        self.assertTrue(MDF)

    @classmethod
    def setUpClass(cls) -> None:
        url = "https://github.com/danielhrisca/asammdf/files/4078993/test.demo.zip"
        urllib.request.urlretrieve(url, "test.zip")

        cls.tempdir_demo = tempfile.TemporaryDirectory()
        cls.tempdir_general = tempfile.TemporaryDirectory()
        cls.tempdir = tempfile.TemporaryDirectory()
        cls.tempdir_array = tempfile.TemporaryDirectory()

        ZipFile(r"test.zip").extractall(cls.tempdir_demo.name)
        Path("test.zip").unlink()
        for version in ("3.30", "4.10"):
            generate_test_file(cls.tempdir_general.name, version)

        generate_arrays_test_file(cls.tempdir_array.name)

    def test_mdf_header(self) -> None:
        mdf = BytesIO(b"M" * 100)
        with self.assertRaises(MdfException):
            MDF(mdf)

        mdf.close()

    def test_wrong_header_version(self) -> None:
        mdf = BytesIO(
            b"MDF     AAAA    amdf500d\x00\x00\x00\x00\x9f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        )
        with self.assertRaises(MdfException):
            MDF(mdf)

        mdf.close()

    def test_read(self) -> None:
        print("MDF read big files")
        for input_file in Path(TestMDF.tempdir_general.name).iterdir():
            print(input_file)

            equal = True

            for inp in (input_file, BytesIO(input_file.read_bytes())):
                with MDF(inp) as mdf:
                    for i, group in enumerate(mdf.groups):
                        v: npt.NDArray[Any]
                        if i == 0:
                            v = np.ones(cycles, dtype=np.uint64)
                            for j in range(1, 20):
                                vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                if not np.array_equal(vals, v * (j - 1)):
                                    equal = False
                        elif i == 1:
                            v = np.ones(cycles, dtype=np.int64)
                            for j in range(1, 20):
                                vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                if not np.array_equal(vals, v * (j - 1) - 0.5):
                                    equal = False
                        elif i == 2:
                            v = np.arange(cycles, dtype=np.int64) / 100.0
                            form = "{} * sin(v)"
                            for j in range(1, 20):
                                vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                f = form.format(j - 1)
                                if not np.array_equal(vals, numexpr.evaluate(f)):
                                    equal = False
                        elif i == 3:
                            v = np.ones(cycles, dtype=np.int64)
                            form = "({} * v -0.5) / 1"
                            for j in range(1, 20):
                                f = form.format(j - 1)
                                vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                if not np.array_equal(vals, numexpr.evaluate(f)):
                                    equal = False
                        elif i == 4:
                            for j in range(1, 20):
                                target = np.array([f"Channel {j} sample {k}".encode("ascii") for k in range(cycles)])
                                vals = mdf.get(group=i, index=j + 1, samples_only=True)[0]
                                cond = np.array_equal(vals, target)
                                if not cond:
                                    print(i, j, vals, target, len(vals), len(target))
                                self.assertTrue(cond)

                        elif i == 5:
                            v = np.ones(cycles, dtype=np.dtype("(8,)u1"))
                            for j in range(1, 20):
                                target = v * j
                                vals = mdf.get(group=i, index=j + 1, samples_only=True)[0]
                                cond = np.array_equal(vals, target)
                                if not cond:
                                    print(i, j, vals, target, len(vals), len(target))
                                self.assertTrue(cond)

                        elif i == 6:
                            for j in range(1, 20):
                                target = np.array([b"Value %d" % j for _ in range(cycles)])
                                vals = mdf.get(group=i, index=j + 1, samples_only=True)[0]
                                cond = np.array_equal(vals, target)
                                if not cond:
                                    print(i, j, vals, target, len(vals), len(target))
                                self.assertTrue(cond)

                if isinstance(inp, BytesIO):
                    inp.close()
            self.assertTrue(equal)

    def test_read_arrays(self) -> None:
        print("MDF read big array files")
        for input_file in Path(TestMDF.tempdir_array.name).iterdir():
            print(input_file)

            equal = True

            for inp in (input_file, BytesIO(input_file.read_bytes())):
                with MDF(inp) as mdf:
                    mdf.configure(read_fragment_size=8000)

                    for i, group in enumerate(mdf.groups):
                        if i == 0:
                            samples_list = [
                                np.ones((cycles, 2, 3), dtype=np.uint64),
                                np.ones((cycles, 2), dtype=np.uint64),
                                np.ones((cycles, 3), dtype=np.uint64),
                            ]

                            for j in range(1, 20):
                                types: list[npt.DTypeLike] = [
                                    (f"Channel_{j}", "(2, 3)<u8"),
                                    (f"channel_{j}_axis_1", "(2, )<u8"),
                                    (f"channel_{j}_axis_2", "(3, )<u8"),
                                ]
                                dtype = np.dtype(types)

                                vals = mdf.get(f"Channel_{j}", group=i, samples_only=True)[0]
                                arrays = [arr * j for arr in samples_list]
                                target = np.rec.fromarrays(arrays, dtype=dtype)
                                if not np.array_equal(vals, target):
                                    equal = False
                                    raise Exception

                        elif i == 1:
                            samples = np.ones((cycles, 2, 3), dtype=np.uint64)
                            axis_0 = np.ones((cycles, 2), dtype=np.uint64)
                            axis_1 = np.ones((cycles, 3), dtype=np.uint64)

                            for j in range(1, 20):
                                types = [(f"Channel_{j}", "(2, 3)<u8")]
                                dtype = np.dtype(types)

                                vals = mdf.get(f"Channel_{j}", group=i, samples_only=True)[0]
                                arrays = [samples * j]
                                target = np.rec.fromarrays(arrays, dtype=dtype)
                                if not np.array_equal(vals, target):
                                    equal = False
                                    raise Exception

                        elif i == 2:
                            samples_list = [
                                np.ones(cycles, dtype=np.uint8),
                                np.ones(cycles, dtype=np.uint16),
                                np.ones(cycles, dtype=np.uint32),
                                np.ones(cycles, dtype=np.uint64),
                                np.ones(cycles, dtype=np.int8),
                                np.ones(cycles, dtype=np.int16),
                                np.ones(cycles, dtype=np.int32),
                                np.ones(cycles, dtype=np.int64),
                            ]

                            for j in range(1, 20):
                                types = [
                                    (f"struct_{j}_channel_0", np.uint8),
                                    (f"struct_{j}_channel_1", np.uint16),
                                    (f"struct_{j}_channel_2", np.uint32),
                                    (f"struct_{j}_channel_3", np.uint64),
                                    (f"struct_{j}_channel_4", np.int8),
                                    (f"struct_{j}_channel_5", np.int16),
                                    (f"struct_{j}_channel_6", np.int32),
                                    (f"struct_{j}_channel_7", np.int64),
                                ]
                                dtype = np.dtype(types)

                                vals = mdf.get(f"Channel_{j}", group=i, samples_only=True)[0]
                                arrays = [arr * j for arr in samples_list]
                                target = np.rec.fromarrays(arrays, dtype=dtype)
                                if not np.array_equal(vals, target):
                                    equal = False
                                    print(target)
                                    print(vals)
                                    raise Exception

                if isinstance(inp, BytesIO):
                    inp.close()

            self.assertTrue(equal)

    def test_read_demo(self) -> None:
        print("MDF read tests")

        mdf_files = [file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix in (".mdf", ".mf4")]

        signals = [file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix == ".npy"]

        for file in mdf_files:
            print(file)

            for inp in (file, BytesIO(file.read_bytes())):
                with MDF(inp, use_display_names=True) as input_file:
                    for signal in signals:
                        name = signal.stem
                        target = np.load(signal)
                        values = input_file.get(name, *input_file.whereis(name)[0]).samples

                        self.assertTrue(np.array_equal(target, values))

                if isinstance(inp, BytesIO):
                    inp.close()

    def test_convert(self) -> None:
        print("MDF convert big files tests")

        for out in SUPPORTED_VERSIONS:
            for input_file in Path(TestMDF.tempdir_general.name).iterdir():
                for compression in (0, 1, 2):
                    print(input_file, out, compression)

                    with MDF(input_file) as mdf:
                        mdf.configure(read_fragment_size=8000)
                        converted = mdf.convert(out)
                        outfile = converted.save(
                            Path(TestMDF.tempdir.name) / f"tmp_convert_{out}",
                            overwrite=True,
                            compression=compression,
                        )
                        converted.close()

                    equal = True

                    with MDF(outfile) as mdf:
                        mdf.configure(read_fragment_size=8000)

                        for i, group in enumerate(mdf.groups):
                            v: npt.NDArray[Any]
                            if i == 0:
                                v = np.ones(cycles, dtype=np.uint64)
                                for j in range(1, 20):
                                    vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                    if not np.array_equal(vals, v * (j - 1)):
                                        equal = False
                                        print(outfile)
                                        print(vals, len(vals))
                                        print(v * (j - 1), len(v))

                                        raise Exception
                            elif i == 1:
                                v = np.ones(cycles, dtype=np.int64)
                                for j in range(1, 20):
                                    vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                    if not np.array_equal(vals, v * (j - 1) - 0.5):
                                        equal = False
                                        raise Exception
                            elif i == 2:
                                v = np.arange(cycles, dtype=np.int64) / 100.0
                                form = "{} * sin(v)"
                                for j in range(1, 20):
                                    vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                    f = form.format(j - 1)
                                    if not np.array_equal(vals, numexpr.evaluate(f)):
                                        equal = False
                                        raise Exception
                            elif i == 3:
                                v = np.ones(cycles, dtype=np.int64)
                                form = "({} * v -0.5) / 1"
                                for j in range(1, 20):
                                    f = form.format(j - 1)
                                    vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                    if not np.array_equal(vals, numexpr.evaluate(f)):
                                        equal = False
                                        target = numexpr.evaluate(f)
                                        print(i, j, vals, target, len(vals), len(target))
                                        raise Exception
                            elif i == 4:
                                for j in range(1, 20):
                                    target = np.array(
                                        [f"Channel {j} sample {k}".encode("ascii") for k in range(cycles)]
                                    )
                                    vals = mdf.get(group=i, index=j + 1, samples_only=True)[0]
                                    cond = np.array_equal(vals, target)
                                    if not cond:
                                        print(i, j, vals, target, len(vals), len(target))
                                        raise Exception
                                    self.assertTrue(cond)

                            elif i == 5:
                                v = np.ones(cycles, dtype=np.dtype("(8,)u1"))
                                for j in range(1, 20):
                                    target = v * j
                                    vals = mdf.get(group=i, index=j + 1, samples_only=True)[0]
                                    cond = np.array_equal(vals, target)
                                    if not cond:
                                        print(i, j, vals, target, len(vals), len(target))
                                        raise Exception
                                    self.assertTrue(cond)

                            elif i == 6:
                                for j in range(1, 20):
                                    target = np.array([b"Value %d" % j for _ in range(cycles)])
                                    vals = mdf.get(group=i, index=j + 1, samples_only=True)[0]
                                    cond = np.array_equal(vals, target)
                                    if not cond:
                                        print(i, j, vals, target, len(vals), len(target))
                                        raise Exception
                                    self.assertTrue(cond)

                    self.assertTrue(equal)

    def test_convert_demo(self) -> None:
        print("MDF convert demo tests")

        mdf_files = [file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix in (".mdf", ".mf4")]

        signals = [file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix == ".npy"]

        for file in mdf_files:
            for inp in (file, BytesIO(file.read_bytes())):
                with MDF(inp, use_display_names=True) as input_file:
                    for out in SUPPORTED_VERSIONS:
                        print(file, out, type(inp))

                    converted = input_file.convert(out)
                    outfile = converted.save(Path(TestMDF.tempdir_demo.name) / "tmp", overwrite=True)
                    converted.close()

                    with MDF(outfile, use_display_names=True) as mdf:
                        for signal in signals:
                            name = signal.stem
                            target = np.load(signal)
                            values = mdf.get(name, *mdf.whereis(name)[0]).samples

                            self.assertTrue(np.array_equal(target, values))

                if isinstance(inp, BytesIO):
                    inp.close()

    def test_cut(self) -> None:
        print("MDF cut big files tests")

        for input_file in Path(TestMDF.tempdir_general.name).iterdir():
            for whence in (0, 1):
                for compression in (0, 1, 2):
                    mdf = MDF(input_file)

                    mdf.configure(read_fragment_size=8000)
                    cut = mdf.cut(stop=-1, whence=whence, include_ends=False)
                    outfile0 = cut.save(Path(TestMDF.tempdir.name) / "tmp0", overwrite=True)
                    cut.close()

                    mdf.configure(read_fragment_size=8000)
                    cut = mdf.cut(stop=105, whence=whence, include_ends=False)
                    outfile1 = cut.save(Path(TestMDF.tempdir.name) / "tmp1", overwrite=True)
                    cut.close()

                    mdf.configure(read_fragment_size=8000)
                    cut = mdf.cut(start=105.1, stop=201, whence=whence, include_ends=False)
                    outfile2 = cut.save(Path(TestMDF.tempdir.name) / "tmp2", overwrite=True)
                    cut.close()

                    mdf.configure(read_fragment_size=8000)
                    cut = mdf.cut(start=201.1, whence=whence, include_ends=False)
                    outfile3 = cut.save(Path(TestMDF.tempdir.name) / "tmp3", overwrite=True)
                    cut.close()

                    mdf.configure(read_fragment_size=8000)
                    cut = mdf.cut(start=7000, whence=whence, include_ends=False)
                    outfile4 = cut.save(Path(TestMDF.tempdir.name) / "tmp4", overwrite=True)
                    cut.close()

                    concatenated = MDF.concatenate(
                        [outfile0, outfile1, outfile2, outfile3, outfile4],
                        version=mdf.version,
                        sync=bool(whence),
                    )

                    outfile = concatenated.save(
                        Path(TestMDF.tempdir.name) / "tmp_cut",
                        overwrite=True,
                        compression=compression,
                    )

                    mdf.close()
                    concatenated.close()

                    with MDF(outfile) as mdf:
                        for i, group in enumerate(mdf.groups):
                            v: npt.NDArray[Any]
                            if i == 0:
                                v = np.ones(cycles, dtype=np.uint64)
                                for j in range(1, 20):
                                    vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                    cond = np.array_equal(vals, v * (j - 1))
                                    if not cond:
                                        print(i, j, vals, v * (j - 1), len(vals), len(v))
                                    self.assertTrue(cond)
                            elif i == 1:
                                v = np.ones(cycles, dtype=np.int64)
                                for j in range(1, 20):
                                    vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                    cond = np.array_equal(vals, v * (j - 1) - 0.5)
                                    if not cond:
                                        print(vals, v * (j - 1) - 0.5, len(vals), len(v))
                                    self.assertTrue(cond)
                            elif i == 2:
                                v = np.arange(cycles, dtype=np.int64) / 100.0
                                form = "{} * sin(v)"
                                for j in range(1, 20):
                                    vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                    f = form.format(j - 1)
                                    target = numexpr.evaluate(f)
                                    cond = np.array_equal(vals, target)
                                    if not cond:
                                        print(i, j, vals, target, len(vals), len(target))
                                    self.assertTrue(cond)
                            elif i == 3:
                                v = np.ones(cycles, dtype=np.int64)
                                form = "({} * v -0.5) / 1"
                                for j in range(1, 20):
                                    f = form.format(j - 1)
                                    vals = mdf.get(group=i, index=j, samples_only=True)[0]
                                    target = numexpr.evaluate(f)
                                    cond = np.array_equal(vals, target)
                                    if not cond:
                                        print(i, j, vals, target, len(vals), len(target))
                                    self.assertTrue(cond)
                            elif i == 4:
                                for j in range(1, 20):
                                    target = np.array(
                                        [f"Channel {j} sample {k}".encode("ascii") for k in range(cycles)]
                                    )
                                    vals = mdf.get(group=i, index=j + 1, samples_only=True)[0]
                                    cond = np.array_equal(vals, target)
                                    if not cond:
                                        print(i, j, vals, target, len(vals), len(target))
                                    self.assertTrue(cond)

                            elif i == 5:
                                v = np.ones(cycles, dtype=np.dtype("(8,)u1"))
                                for j in range(1, 20):
                                    target = v * j
                                    vals = mdf.get(group=i, index=j + 1, samples_only=True)[0]
                                    cond = np.array_equal(vals, target)
                                    if not cond:
                                        print(i, j, vals, target, len(vals), len(target))
                                    self.assertTrue(cond)

                            elif i == 6:
                                for j in range(1, 20):
                                    target = np.array([b"Value %d" % j for _ in range(cycles)])
                                    vals = mdf.get(group=i, index=j + 1, samples_only=True)[0]
                                    cond = np.array_equal(vals, target)
                                    if not cond:
                                        print(i, j, vals, target, len(vals), len(target))
                                    self.assertTrue(cond)

    def test_cut_arrays(self) -> None:
        print("MDF cut big array files")
        for input_file in Path(TestMDF.tempdir_array.name).iterdir():
            for whence in (0, 1):
                print(input_file, whence)

                mdf = MDF(input_file)

                mdf.configure(read_fragment_size=8000)
                cut = mdf.cut(stop=105.5, whence=whence, include_ends=False)
                outfile1 = cut.save(Path(TestMDF.tempdir.name) / "tmp1", overwrite=True)
                cut.close()

                cut = mdf.cut(start=105.5, stop=201.5, whence=whence, include_ends=False)
                outfile2 = cut.save(Path(TestMDF.tempdir.name) / "tmp2", overwrite=True)
                cut.close()

                cut = mdf.cut(start=201.5, whence=whence, include_ends=False)
                outfile3 = cut.save(Path(TestMDF.tempdir.name) / "tmp3", overwrite=True)
                cut.close()

                concatenated = MDF.concatenate([outfile1, outfile2, outfile3], mdf.version)
                outfile = concatenated.save(Path(TestMDF.tempdir.name) / "tmp_cut", overwrite=True)

                mdf.close()
                concatenated.close()

                equal = True

                with MDF(outfile) as mdf:
                    mdf.configure(read_fragment_size=8000)

                    for i, group in enumerate(mdf.groups):
                        if i == 0:
                            samples_list = [
                                np.ones((cycles, 2, 3), dtype=np.uint64),
                                np.ones((cycles, 2), dtype=np.uint64),
                                np.ones((cycles, 3), dtype=np.uint64),
                            ]

                            for j in range(1, 20):
                                types: list[npt.DTypeLike] = [
                                    (f"Channel_{j}", "(2, 3)<u8"),
                                    (f"channel_{j}_axis_1", "(2, )<u8"),
                                    (f"channel_{j}_axis_2", "(3, )<u8"),
                                ]
                                dtype = np.dtype(types)

                                vals = mdf.get(f"Channel_{j}", group=i, samples_only=True)[0]
                                arrays = [arr * j for arr in samples_list]
                                target = np.rec.fromarrays(arrays, dtype=dtype)
                                if not np.array_equal(vals, target):
                                    equal = False
                                    print(
                                        i,
                                        j,
                                        len(target),
                                        len(vals),
                                        vals,
                                        target,
                                        sep="\n\n",
                                    )
                                    raise Exception

                        elif i == 1:
                            samples = np.ones((cycles, 2, 3), dtype=np.uint64)
                            axis_0 = np.ones((cycles, 2), dtype=np.uint64)
                            axis_1 = np.ones((cycles, 3), dtype=np.uint64)

                            for j in range(1, 20):
                                types = [(f"Channel_{j}", "(2, 3)<u8")]
                                dtype = np.dtype(types)

                                vals = mdf.get(f"Channel_{j}", group=i, samples_only=True)[0]
                                arrays = [samples * j]
                                target = np.rec.fromarrays(arrays, dtype=dtype)
                                if not np.array_equal(vals, target):
                                    equal = False
                                    raise Exception

                        elif i == 2:
                            samples_list = [
                                np.ones(cycles, dtype=np.uint8),
                                np.ones(cycles, dtype=np.uint16),
                                np.ones(cycles, dtype=np.uint32),
                                np.ones(cycles, dtype=np.uint64),
                                np.ones(cycles, dtype=np.int8),
                                np.ones(cycles, dtype=np.int16),
                                np.ones(cycles, dtype=np.int32),
                                np.ones(cycles, dtype=np.int64),
                            ]

                            for j in range(1, 20):
                                types = [
                                    (f"struct_{j}_channel_0", np.uint8),
                                    (f"struct_{j}_channel_1", np.uint16),
                                    (f"struct_{j}_channel_2", np.uint32),
                                    (f"struct_{j}_channel_3", np.uint64),
                                    (f"struct_{j}_channel_4", np.int8),
                                    (f"struct_{j}_channel_5", np.int16),
                                    (f"struct_{j}_channel_6", np.int32),
                                    (f"struct_{j}_channel_7", np.int64),
                                ]
                                dtype = np.dtype(types)

                                vals = mdf.get(f"Channel_{j}", group=i, samples_only=True)[0]
                                arrays = [arr * j for arr in samples_list]
                                target = np.rec.fromarrays(arrays, dtype=dtype)
                                if not np.array_equal(vals, target):
                                    equal = False
                                    raise Exception

            self.assertTrue(equal)

    def test_cut_demo(self) -> None:
        print("MDF cut demo tests")

        mdf_files = [file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix in (".mdf", ".mf4")]

        signals = [file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix == ".npy"]

        for file in mdf_files:
            print(file)

            for inp in (file, BytesIO(file.read_bytes())):
                with MDF(inp, use_display_names=True) as input_file:
                    for whence in (0, 1):
                        print(file, whence)

                        cut = input_file.cut(stop=2, whence=whence, include_ends=False)
                        outfile1 = cut.save(Path(TestMDF.tempdir.name) / "tmp1", overwrite=True)
                        cut.close()

                        cut = input_file.cut(start=2, stop=6, whence=whence, include_ends=False)
                        outfile2 = cut.save(Path(TestMDF.tempdir.name) / "tmp2", overwrite=True)
                        cut.close()

                        cut = input_file.cut(start=6, whence=whence, include_ends=False)
                        outfile3 = cut.save(Path(TestMDF.tempdir.name) / "tmp3", overwrite=True)
                        cut.close()

                        concatenated = MDF.concatenate(
                            [outfile1, outfile2, outfile3],
                            version=input_file.version,
                            use_display_names=True,
                        )

                        outfile = concatenated.save(Path(TestMDF.tempdir.name) / "tmp", overwrite=True)

                        print("OUT", outfile)
                        concatenated.close()

                        with MDF(outfile, use_display_names=True) as mdf2:
                            for signal in signals:
                                target = np.load(signal)
                                sig = mdf2.get(signal.stem, *mdf2.whereis(signal.stem)[0])
                                timestamps = input_file.get(signal.stem, *input_file.whereis(signal.stem)[0]).timestamps

                                self.assertTrue(np.array_equal(sig.samples, target))
                                self.assertTrue(np.array_equal(timestamps, sig.timestamps))

                if isinstance(inp, BytesIO):
                    inp.close()

    def test_filter(self) -> None:
        print("MDF read tests")

        mdf_files = [file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix in (".mdf", ".mf4")]

        signals = {file.stem: file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix == ".npy"}

        for file in mdf_files:
            print(file)

            for inp in (file, BytesIO(file.read_bytes())):
                with MDF(inp, use_display_names=True) as input_file:
                    names = [ch.name for gp in input_file.groups for ch in gp.channels[1:]]

                    channels_nr = np.random.randint(1, len(names) + 1)

                    channel_list = random.sample(names, channels_nr)

                    filtered_mdf = input_file.filter(channel_list)

                    target_names = set(filtered_mdf.channels_db)

                    self.assertFalse(set(channel_list) - (target_names - {"time"}))

                    for name in channel_list:
                        target = np.load(signals[name])
                        filtered = filtered_mdf.get(name)
                        self.assertTrue(np.array_equal(target, filtered.samples))

                    filtered_mdf.close()

                if isinstance(inp, BytesIO):
                    inp.close()

    def test_select(self) -> None:
        print("MDF select tests")

        mdf_files = [file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix in (".mdf", ".mf4")]

        signals = {file.stem: file for file in Path(TestMDF.tempdir_demo.name).iterdir() if file.suffix == ".npy"}

        for file in mdf_files:
            print(file)

            for inp in (file, BytesIO(file.read_bytes())):
                with MDF(inp, use_display_names=True) as input_file:
                    names = [ch.name for gp in input_file.groups for ch in gp.channels[1:]]

                    input_file.configure(read_fragment_size=200)

                    channels_nr = np.random.randint(1, len(names) + 1)
                    channel_list = random.sample(names, channels_nr)

                    selected_signals = input_file.select(channel_list)

                    target_names = {s.name for s in selected_signals}

                    self.assertFalse(set(target_names) - set(channel_list))

                    for name, sig in zip(channel_list, selected_signals, strict=False):
                        target = np.load(signals[name])
                        self.assertTrue(np.array_equal(target, sig.samples))

                if isinstance(inp, BytesIO):
                    inp.close()

    def test_scramble(self) -> None:
        print("MDF scramble tests")

        for input_file in Path(TestMDF.tempdir_demo.name).iterdir():
            if input_file.suffix.lower() in (".mdf", ".mf4"):
                scrambled = MDF.scramble(input_file)
                self.assertTrue(scrambled)
                Path(scrambled).unlink()

    def test_iter_groups(self) -> None:
        dfs = [
            DataFrame(
                {
                    f"df_{i}_column_0": np.ones(5) * i,
                    f"df_{i}_column_1": np.arange(5) * i,
                }
            )
            for i in range(5)
        ]

        mdf = MDF()
        for df in dfs:
            mdf.append(df)

        for i, mdf_df in enumerate(mdf.iter_groups()):
            self.assertTrue(mdf_df.equals(dfs[i]))

        mdf.close()

    def test_resample_raster_0(self) -> None:
        sigs = [
            Signal(
                samples=np.ones(1000) * i,
                timestamps=np.arange(1000),
                name=f"Signal_{i}",
            )
            for i in range(20)
        ]

        mdf = MDF()
        mdf.append(sigs)
        mdf.configure(read_fragment_size=1)
        with self.assertRaises(MdfException):
            mdf = mdf.resample(raster=0)

        mdf.close()

    def test_resample(self) -> None:
        raster = 1.33
        sigs = [
            Signal(
                samples=np.arange(1000, dtype="f8"),
                timestamps=np.concatenate([np.arange(500), np.arange(1000, 1500)]),
                name=f"Signal_{i}",
            )
            for i in range(20)
        ]

        mdf = MDF()
        mdf.append(sigs)
        resampled = mdf.resample(raster=raster)
        mdf.close()

        target_timestamps = np.arange(0, 1500, 1.33)
        target_samples = np.concatenate(
            [
                np.arange(0, 500, 1.33),
                np.linspace(499.00215568862274, 499.9976646706587, 376),
                np.arange(500.1600000000001, 1000, 1.33),
            ]
        )

        for i, sig in enumerate(resampled.iter_channels(skip_master=True)):
            self.assertTrue(np.array_equal(sig.timestamps, target_timestamps))
            self.assertTrue(np.allclose(sig.samples, target_samples))

        resampled.close()

    def test_to_dataframe(self) -> None:
        dfs = [
            DataFrame(
                {
                    f"df_{i}_column_0": np.ones(5) * i,
                    f"df_{i}_column_1": np.arange(5) * i,
                }
            )
            for i in range(5)
        ]

        mdf = MDF()
        for df in dfs:
            mdf.append(df)

        target: dict[str, npt.NDArray[Any]] = {}
        for i in range(5):
            target[f"df_{i}_column_0"] = np.ones(5) * i
            target[f"df_{i}_column_1"] = np.arange(5) * i

        df = DataFrame(target)

        self.assertTrue(df.equals(mdf.to_dataframe()))
        mdf.close()

    def test_search(self) -> None:
        sigs = [
            Signal(
                samples=np.ones(1),
                timestamps=np.arange(1),
                name=ch_name,
            )
            for ch_name in ["foo", "Bar", "baz"]
        ]
        mdf = MDF()
        mdf.append(sigs)

        # plain text matching
        self.assertEqual(["foo"], mdf.search(pattern="foo"), msg="name match full")
        self.assertEqual(["foo"], mdf.search(pattern="oo", mode="plain"), msg="name match part")
        self.assertEqual(
            [],
            mdf.search(pattern="FOO", mode=SearchMode.plain, case_insensitive=False),
            msg="name match case-sensitive (no match)",
        )
        self.assertEqual(
            ["foo"],
            mdf.search(pattern="FOO", mode=SearchMode.plain, case_insensitive=True),
            msg="name match case-insensitive (match)",
        )

        # regex matching
        self.assertEqual(["Bar"], mdf.search(pattern="^Bar$", mode="regex"), msg="regex match full")
        self.assertEqual(
            ["baz"],
            mdf.search(pattern="z$", mode=SearchMode.regex),
            msg="regex match part",
        )
        self.assertEqual(
            ["baz"],
            mdf.search(pattern="b.*", mode=SearchMode.regex, case_insensitive=False),
            msg="regex match case-sensitive",
        )
        self.assertEqual(
            ["Bar", "baz"],
            mdf.search(pattern="b.*", mode=SearchMode.regex, case_insensitive=True),
            msg="regex match case-insensitive",
        )

        # wildcard matching
        self.assertEqual(
            ["Bar"],
            mdf.search(pattern="Bar", mode="wildcard"),
            msg="wildcard match full",
        )
        self.assertEqual(
            ["foo"],
            mdf.search(pattern="*oo", mode=SearchMode.wildcard),
            msg="wildcard match part",
        )
        self.assertEqual(
            ["baz"],
            mdf.search(pattern="b*", mode=SearchMode.wildcard, case_insensitive=False),
            msg="wildcard match case-sensitive",
        )
        self.assertEqual(
            ["Bar", "baz"],
            mdf.search(pattern="b*", mode=SearchMode.wildcard, case_insensitive=True),
            msg="wildcard match case-insensitive",
        )


if __name__ == "__main__":
    unittest.main()



from collections.abc import Sequence
from os import PathLike
from typing import Literal, TYPE_CHECKING, Union

from canmatrix import CanMatrix
from numpy.typing import NDArray
from typing_extensions import Any

if TYPE_CHECKING:
    from . import v2_v3_blocks as v3b
    from . import v4_blocks as v4b
    from . import v4_constants as v4c
    from .source_utils import Source


StrPath = str | PathLike[str]

# asammdf specific types

BusType = Literal["CAN", "LIN"]
ChannelConversionType = Union["v3b.ChannelConversion", "v4b.ChannelConversion"]
ChannelsType = Sequence[str | tuple[str | None, int, int] | tuple[str, int]]
CompressionType = Union[Literal[0, 1, 2, 3, 4, 5], "v4c.CompressionAlgorithm"]
DbcFileType = tuple[StrPath | CanMatrix, int]
EmptyChannelsType = Literal["skip", "zeros"]
FloatInterpolationModeType = Literal[0, 1]
IntInterpolationModeType = Literal[0, 1, 2]
RasterType = float | str | NDArray[Any]
SourceType = Union["v3b.ChannelExtension", "v4b.SourceInformation", "Source"]
